from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    Line,
    ManualMainController,
    MicrogridMode,
    Production,
)
from relsad.network.systems import (
    Distribution,
    Microgrid,
    PowerSystem,
    Transmission,
)
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network(
    fail_rate_trafo: float = 0.007,
    fail_rate_line: float = 0.7,
    microgrid_mode: MicrogridMode = MicrogridMode.SURVIVAL,
    line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
):

    C1 = ManualMainController(
        name="C1",
        sectioning_time=Time(1, TimeUnit.HOUR),
    )

    ps = PowerSystem(C1)

    B1 = Bus(
        name="B1",
        coordinate=[0, 0],
        fail_rate_per_year=0,
    )

    B2 = Bus(
        name="B2",
        coordinate=[1, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        name="B3",
        coordinate=[2, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        name="B4",
        coordinate=[3, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        name="B5",
        coordinate=[4, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        name="B6",
        coordinate=[5, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B7 = Bus(
        name="B7",
        coordinate=[6, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B8 = Bus(
        name="B8",
        coordinate=[7, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B9 = Bus(
        name="B9",
        coordinate=[8, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B10 = Bus(
        name="B10",
        coordinate=[9, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B11 = Bus(
        name="B11",
        coordinate=[10, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B12 = Bus(
        name="B12",
        coordinate=[11, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B13 = Bus(
        name="B13",
        coordinate=[12, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B14 = Bus(
        name="B14",
        coordinate=[13, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B15 = Bus(
        name="B15",
        coordinate=[14, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B16 = Bus(
        name="B16",
        coordinate=[15, 0],
        fail_rate_per_year=fail_rate_trafo,
    )

    # Microgrid:
    BM1 = Bus(
        name="BM1",
        coordinate=[13, 1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    BM2 = Bus(
        name="BM2",
        coordinate=[14, 1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    BM3 = Bus(
        name="BM3",
        coordinate=[14, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    BM4 = Bus(
        name="BM4",
        coordinate=[14, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    Battery("Bat1", BM1)
    Production("P1", BM3)
    Production("P2", BM4)

    # Lines, connections and impedances
    L1 = Line(
        name="L1",
        fbus=B1,
        tbus=B2,
        r=0.0922,  # 0.3660,
        x=0.0470,  # 0.1864,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L2 = Line(
        name="L2",
        fbus=B2,
        tbus=B3,
        r=0.0922,  # 0.3811,
        x=0.0470,  # 0.1941,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L3 = Line(
        name="L3",
        fbus=B3,
        tbus=B4,
        r=0.0922,
        x=0.0470,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L4 = Line(
        name="L4",
        fbus=B4,
        tbus=B5,
        r=0.0493,
        x=0.0251,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L5 = Line(
        name="L5",
        fbus=B5,
        tbus=B6,
        r=0.1872,  # 0.8190,
        x=0.0619,  # 0.2707,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L6 = Line(
        name="L6",
        fbus=B6,
        tbus=B7,
        r=0.1872,
        x=0.0619,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L7 = Line(
        name="L7",
        fbus=B7,
        tbus=B8,
        r=0.1872,  # 0.7114,
        x=0.0619,  # 0.2351,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L8 = Line(
        name="L8",
        fbus=B8,
        tbus=B9,
        r=0.1872,  # 1.0300,
        x=0.0619,  # 0.3400,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L9 = Line(
        name="L9",
        fbus=B9,
        tbus=B10,
        r=0.1872,  # 1.0440,
        x=0.0619,  # 0.3450,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L10 = Line(
        name="L10",
        fbus=B10,
        tbus=B11,
        r=0.1872,  # 1.0580,
        x=0.0619,  # 0.3496,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L11 = Line(
        name="L11",
        fbus=B11,
        tbus=B12,
        r=0.1966,
        x=0.0650,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L12 = Line(
        name="L12",
        fbus=B12,
        tbus=B13,
        r=0.1872,  # 1.0300,
        x=0.0619,  # 0.3400,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L13 = Line(
        name="L13",
        fbus=B13,
        tbus=B14,
        r=0.1872,  # 1.0440,
        x=0.0619,  # 0.3450,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L14 = Line(
        name="L14",
        fbus=B14,
        tbus=B15,
        r=0.1872,  # 1.0580,
        x=0.0619,  # 0.3496,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L15 = Line(
        name="L15",
        fbus=B15,
        tbus=B16,
        r=0.1966,
        x=0.0650,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )

    # Microgrid

    ML1 = Line(
        name="ML1",
        fbus=B13,
        tbus=BM1,
        r=0.1872,  # 0.7394,
        x=0.0619,  # 0.2444,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )

    ML2 = Line(
        name="ML2",
        fbus=BM1,
        tbus=BM2,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    ML3 = Line(
        name="ML3",
        fbus=BM1,
        tbus=BM3,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    ML4 = Line(
        name="ML4",
        fbus=BM1,
        tbus=BM4,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )

    CircuitBreaker(name="E1", line=L1)
    CircuitBreaker(name="E2", line=ML1)

    Disconnector(name="L1a", line=L1, bus=B1)
    Disconnector(name="L1b", line=L1, bus=B2)
    Disconnector(name="L2a", line=L2, bus=B2)
    Disconnector(name="L2b", line=L2, bus=B3)
    Disconnector(name="L3a", line=L3, bus=B3)
    Disconnector(name="L3b", line=L3, bus=B4)
    Disconnector(name="L4a", line=L4, bus=B4)
    Disconnector(name="L4b", line=L4, bus=B5)
    Disconnector(name="L5a", line=L5, bus=B5)
    Disconnector(name="L5b", line=L5, bus=B6)
    Disconnector(name="L6a", line=L6, bus=B6)
    Disconnector(name="L6b", line=L6, bus=B7)
    Disconnector(name="L7a", line=L7, bus=B7)
    Disconnector(name="L7b", line=L7, bus=B8)
    Disconnector(name="L8a", line=L8, bus=B8)
    Disconnector(name="L8b", line=L8, bus=B9)
    Disconnector(name="L9a", line=L9, bus=B9)
    Disconnector(name="L9b", line=L9, bus=B10)
    Disconnector(name="L10a", line=L10, bus=B10)
    Disconnector(name="L10b", line=L10, bus=B11)
    Disconnector(name="L11a", line=L11, bus=B11)
    Disconnector(name="L11b", line=L11, bus=B12)
    Disconnector(name="L12a", line=L12, bus=B12)
    Disconnector(name="L12b", line=L12, bus=B13)
    Disconnector(name="L13a", line=L13, bus=B13)
    Disconnector(name="L13b", line=L13, bus=B14)
    Disconnector(name="L14a", line=L14, bus=B14)
    Disconnector(name="L14b", line=L14, bus=B15)
    Disconnector(name="L15a", line=L15, bus=B15)
    Disconnector(name="L15b", line=L15, bus=B16)

    # Microgrid:

    Disconnector(name="ML1a", line=ML1, bus=B13)
    Disconnector(name="ML1b", line=ML1, bus=BM1)
    Disconnector(name="ML2a", line=ML2, bus=BM1)
    Disconnector(name="ML2b", line=ML2, bus=BM2)
    Disconnector(name="ML3a", line=ML3, bus=BM1)
    Disconnector(name="ML3b", line=ML3, bus=BM3)
    Disconnector(name="ML4a", line=ML4, bus=BM1)
    Disconnector(name="ML4b", line=ML4, bus=BM4)

    tn = Transmission(ps, B1)

    dn = Distribution(tn, L1)

    dn.add_buses(
        [
            B2,
            B3,
            B4,
            B5,
            B6,
            B7,
            B8,
            B9,
            B10,
            B11,
            B12,
            B13,
            B14,
            B15,
            B16,
        ]
    )

    dn.add_lines(
        [
            L2,
            L3,
            L4,
            L5,
            L6,
            L7,
            L8,
            L9,
            L10,
            L11,
            L12,
            L13,
            L14,
            L15,
        ]
    )

    m = Microgrid(dn, ML1, mode=microgrid_mode)
    m.add_buses([BM1, BM2, BM3, BM4])
    m.add_lines([ML2, ML3, ML4])

    return ps


if __name__ == "__main__":
    import os

    ps = initialize_network()
    fig = plot_topology(ps.buses, ps.lines)

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "IEEE16_modified.pdf"
        )
    )



import os

import matplotlib.pyplot as plt
import pandas as pd

from relsad.Time import TimeUnit


def plot_history(
    obj_list: list,
    attribute: str,
    time_unit: TimeUnit,
    save_dir: str,
):
    """
    Plots the history

    Parameters
    ----------
    obj_list : list
        List of objects
    attribute : str
        An attribute
    time_unit : TimeUnit
        The time unit of the simulation
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)
    fig = plt.figure(dpi=150)
    ax = fig.add_subplot(1, 1, 1)
    for obj in obj_list:
        data = obj.get_history(attribute)
        ax.plot(list(data.values()), label=obj.name)
    ax.set_title(attribute)
    ax.set_xlabel(time_unit.name)
    ax.legend()
    fig.savefig(os.path.join(save_dir, attribute + ".pdf"), format="pdf")
    plt.close(fig)


def plot_monte_carlo_history(
    obj_list: list,
    attribute: str,
    save_dir: str,
):
    """
    Plots the history from the Monte Carlo simulation

    Parameters
    ----------
    obj_list : list
        List of objects
    attribute : str
        An attribute
    save_dir : str
        The saving directory

    Returns
    ----------
    None


    """
    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)
    fig = plt.figure(dpi=150)
    ax = fig.add_subplot(1, 1, 1)
    for obj in obj_list:
        data = obj.get_monte_carlo_history(attribute)
        ax.plot(list(data.values()), label=obj.name)
    ax.set_title(attribute)
    ax.legend()
    fig.savefig(os.path.join(save_dir, attribute + ".pdf"), format="pdf")
    plt.close(fig)


def plot_history_last_state(
    obj_list: list,
    attribute: str,
    save_dir: str,
):
    """
    Plots the last state from the history

    Parameters
    ----------
    obj_list : list
        List of objects
    attribute : str
        An attribute
    save_dir : str
        The saving directory

    Returns
    ----------
    None


    """
    if not os.path.isdir(save_dir):
        os.mkdir(save_dir)
    fig = plt.figure(dpi=150)
    ax = fig.add_subplot(1, 1, 1)
    df = pd.DataFrame()
    for obj in obj_list:
        data = obj.get_history(attribute)
        df[obj] = data.values()
    df.iloc[-1].plot.bar()
    ax.set_title(attribute)
    ax.legend()
    fig.savefig(os.path.join(save_dir, attribute + ".pdf"), format="pdf")
    plt.close(fig)



import os

from relsad.network.components import MainController
from relsad.network.systems import PowerSystem
from relsad.results.storage import save_history
from relsad.Time import TimeUnit
from relsad.visualization.plotting import plot_history


def plot_obj_history(
    obj_list: list,
    time_unit: TimeUnit,
    save_dir: str,
):
    """
    Plots the history of a list of objects in the power system

    Parameters
    ----------
    obj_list : list
        List of objects
    time_unit : TimeUnit
        The time unit of the simulation
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    for attribute in obj_list[0].history:
        plot_history(
            obj_list=obj_list,
            time_unit=time_unit,
            attribute=attribute,
            save_dir=save_dir)


def save_obj_history(
    obj_list: list,
    time_unit: TimeUnit,
    save_dir: str,
):
    """
    Saves the history of a list of objects in the power system

    Parameters
    ----------
    obj_list : list
        List of objects
    time_unit : TimeUnit
        The time unit of the simulation
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    for attribute in obj_list[0].history:
        save_history(
            obj_list=obj_list,
            attribute=attribute,
            time_unit=time_unit,
            save_dir=save_dir,
        )


def save_sequence_history(
    power_system: PowerSystem,
    time_unit: TimeUnit,
    save_dir: str,
):
    """
    Saves the history from an sequence

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    time_unit : TimeUnit
        The time unit of the simulation
    save_dir : str
        The saving path

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    # Power system
    save_obj_history(
        obj_list=[power_system],
        time_unit=time_unit,
        save_dir=os.path.join(save_dir, power_system.name),
    )
    # Child networks
    for network in power_system.child_network_list:
        save_obj_history(
            obj_list=[network],
            time_unit=time_unit,
            save_dir=os.path.join(save_dir, network.name),
        )
    # Buses
    comp_save_dict = {
        "bus": power_system.buses,
        "ev_parks": power_system.ev_parks,
        "battery": power_system.batteries,
        "line": power_system.lines,
        "circuitbreaker": power_system.circuitbreakers,
        "disconnector": power_system.disconnectors,
        "intelligent_switch": power_system.intelligent_switches,
        "sensor": power_system.sensors,
        "distribution_controllers": power_system.controller.distribution_controllers,
        "microgrid_controllers": power_system.controller.microgrid_controllers,
        "main_controller": (
            [power_system.controller]
            if isinstance(power_system.controller, MainController)
            else []
        ),
        "ict_line": power_system.ict_lines,
        "ict_node": power_system.ict_nodes,
    }

    for obj_name, obj_list in comp_save_dict.items():
        if len(obj_list) > 0:
            save_obj_history(
                obj_list=obj_list,
                time_unit=time_unit,
                save_dir=os.path.join(save_dir, obj_name),
            )



from abc import ABC, abstractmethod
from enum import Enum

from relsad.Time import Time


class ControllerState(Enum):
    """
    Controller state

    Attributes
    ----------
    OK : int
        The controller is up and running
    SOFTWARE_FAIL : int
        A software fail has occurred
    HARDWARE_FAIL : int
        A hardware fail has occurred
    REPAIR : int
        The controller is being repaired
    """

    OK = 1
    SOFTWARE_FAIL = 2
    HARDWARE_FAIL = 3
    REPAIR = 4


class Controller(ABC):
    @abstractmethod
    def run_control_loop(self, curr_time: Time, dt: Time):
        pass



from relsad.network.systems import PowerNetwork


def EV_Index(network: PowerNetwork):
    """
    Returns the current EV Index, which is the current power demand from
    all the EV parks in the system.

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_index : float
        Index reflecting the current power demand from all the EV parks in
        the system

    """
    ev_index = sum(ev_park.get_ev_index() for ev_park in network.ev_parks)
    return ev_index


def EV_Interruption(network: PowerNetwork):
    """
    Returns the current EV Interruption. Reflects the average number of
    interruptions per EV car for grid support

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_interruption : float
        The average number of interruptions per EV car for grid support

    """
    interruptions_x_num_cars = sum(
        ev_park.acc_exp_interruptions * ev_park.num_cars
        for ev_park in network.ev_parks
    )
    total_num_cars = sum(ev_park.num_cars for ev_park in network.ev_parks)
    if total_num_cars == 0:
        ev_interruption = 0
    else:
        ev_interruption = interruptions_x_num_cars / total_num_cars
    return ev_interruption


def EV_Duration(network: PowerNetwork):
    """
    Returns the current EV Duration. Reflects the average duration of an EV car
    interruption for grid support

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_duration : float
        The average duration of an EV car interruption for grid support

    """
    sum_interruption_duration = sum(
        ev_park.acc_interruption_duration.get_hours()
        for ev_park in network.ev_parks
    )
    sum_interruptions = sum(
        ev_park.acc_num_interruptions for ev_park in network.ev_parks
    )
    if sum_interruptions == 0:
        ev_duration = 0
    else:
        ev_duration = sum_interruption_duration / sum_interruptions
    return ev_duration



import matplotlib.lines as mlines
import numpy as np

from relsad.StatDist import StatDist, StatDistType, UniformParameters
from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Component import Component
from .ICTNode import ICTNode


class ICTLine(Component):
    r"""
    A class used to represent an ICT Line

    ...

    Attributes
    ----------
    name : str
        Name of the line
    fnode : ICTNode
        Sending ICT node
    tnode : ICTNode
        Receiving ICT node
    parent_network : ICTNetwork
        The parent ICT network of the line
    fail_rate_per_year : float
        Failure rate per year [fault/year/km]
    repair_time_dist : StatDist
        The repair time of the line [hours/fault]
    connected : bool
        Indicates if the line is connected or disconnected
    failed : bool
        Failure status of the line
    remaining_outage_time : Time
        The remaining outage time of the line
    history : dict
        Dictonary attribute that stores the historic variables


    Methods
    -------
    disconnect()
        Disconnects a line and removes the line for the list of lines
    connect()
        Connects a line and append the line to the list of lines
    draw_repair_time(dt)
        Decides and returns the repair time of the line based on a statistical distribution
    fail(dt)
        Sets the fail status of the line to True and opens the connected disconnectors and the connected circuit breaker
    not_fail()
        Sets the fail status of the line to False and closes the connected disconnectors and connected circuit breaker
    change_direction()
        Changes the direction of the line
    update_fail_status(dt)
        Updates the fail status of the line
    print_status()
        Prints the line status
    add_parent_network(network)
        Adds the parent network to the line
    initialize_history()
        Initialize the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    reset_status()
        Resets and sets the status of the class parameters
    """
    lineCount = 0

    ## Visual attributes
    linestyle = "dotted"
    handle = mlines.Line2D([], [], linestyle=linestyle)

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        fnode: ICTNode,
        tnode: ICTNode,
        repair_time_dist: StatDist = StatDist(
            stat_dist_type=StatDistType.UNIFORM_FLOAT,
            parameters=UniformParameters(
                min_val=0.0,
                max_val=0.0,
            ),
        ),
        fail_rate_per_year: float = 0.0,  # fails/year
        connected=True,
    ):

        # Verify input
        if fnode is None:
            raise Exception("Line must have a origin ICT node")
        if tnode is None:
            raise Exception("Line must have a target ICT node")
        if repair_time_dist is None:
            raise Exception("The line must have a repair time distribution")
        if fail_rate_per_year < 0:
            raise Exception("The failure rate per year must be positive")

        ## Informative attributes
        self.name = name

        ## Topological attributes
        self.fnode = fnode
        self.tnode = tnode
        fnode.connected_lines.append(self)
        tnode.connected_lines.append(self)
        tnode.toline = self
        tnode.toline_list.append(self)
        fnode.fromline = self
        fnode.fromline_list.append(self)
        self.parent_network = None
        ICTLine.lineCount += 1

        ## Reliabilility attributes
        self.fail_rate_per_year = fail_rate_per_year
        self.repair_time_dist = repair_time_dist

        ## Status attribute
        self.connected = connected
        self.failed = False
        self.remaining_outage_time = Time(0)

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"ICTLine(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, ICTLine)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def disconnect(self):
        """
        Disconnects a line and removes the line from the list of lines

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if self.connected:
            self.connected = False
            self.linestyle = "--"
            self.fnode.fromline_list.remove(self)
            if self.fnode.fromline == self:
                if len(self.fnode.fromline_list) > 0:
                    self.fnode.fromline = next(iter(self.fnode.fromline_list))
                else:
                    self.fnode.fromline = None
            self.tnode.toline_list.remove(self)
            if self.tnode.toline == self:
                if len(self.tnode.toline_list) > 0:
                    self.tnode.toline = next(iter(self.tnode.toline_list))
                else:
                    self.tnode.toline = None

    def connect(self):
        """
        Connects a line and append the line to the list of lines

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if not self.connected:
            self.connected = True
            self.linestyle = "-"
            self.tnode.toline = self
            self.tnode.toline_list.append(self)
            self.fnode.fromline = self
            self.fnode.fromline_list.append(self)

    def draw_repair_time(self, dt: Time):
        """
        Decides and returns the repair time of the line based on a statistical distribution

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        return Time(
            self.repair_time_dist.draw(
                random_instance=self.ps_random,
                size=1,
            )[0],
            dt.unit,
        )

    def fail(self, dt: Time):
        """
        Sets the fail status of the line to False

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.failed = True
        self.parent_network.failed_line = True
        self.remaining_outage_time = self.draw_repair_time(dt)
        self.disconnect()

    def not_fail(self):
        """
        Sets the fail status of the line to False

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if (
            sum([line.failed for line in self.parent_network.lines]) == 1
            and self.failed
        ):
            self.parent_network.failed_line = False
        self.failed = False
        self.connect()

    def change_direction(self):
        """
        Changes the direction of the line

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.fnode.fromline_list.remove(self)
        self.tnode.fromline_list.append(self)
        self.tnode.toline_list.remove(self)
        self.fnode.toline_list.append(self)
        if self.fnode.fromline == self:
            self.fnode.fromline = (
                next(iter(self.fnode.fromline_list))
                if len(self.fnode.fromline_list) > 0
                else None
            )
        if self.tnode.toline == self:
            self.tnode.toline = (
                next(iter(self.tnode.toline_list))
                if len(self.tnode.toline_list) > 0
                else None
            )
        self.fnode.toline = self
        self.tnode.fromline = self
        node = self.fnode
        self.fnode = self.tnode
        self.tnode = node

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the line

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.failed:
            self.remaining_outage_time -= dt
            if self.remaining_outage_time <= Time(0):
                self.not_fail()
                self.remaining_outage_time = Time(0)
        else:
            p_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
            if random_choice(self.ps_random, p_fail):
                self.fail(dt)
            else:
                self.not_fail()

    def print_status(self):
        """
        Prints the line status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        print(
            "name: {:5s}, failed={}, connected={}".format(
                self.name, self.failed, self.connected
            )
        )

    def add_parent_network(self, network):
        """
        Adds the parent network to the line

        Parameters
        ----------
        network : PowerNetwork
            The parent network of the line

        Returns
        ----------
        None

        """

        self.parent_network = network

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["remaining_outage_time"] = {}
        self.history["failed"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            The vurrent time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["remaining_outage_time"][
                time
            ] = self.remaining_outage_time.get_unit_quantity(curr_time.unit)
            self.history["failed"][time] = self.failed

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.remaining_outage_time = Time(0)

        self.not_fail()
        if save_flag:
            self.initialize_history()


if __name__ == "__main__":
    pass



import os
import sys

from relsad.examples.load.load_and_gen_data import (
    load_data,
    pv_power,
    weather_generation_data,
    wind_power,
)
from relsad.load.bus import CostFunction
from relsad.network.systems import PowerSystem
from relsad.Time import Time, TimeStamp, TimeUnit


def set_network_load_and_prod(
    power_system: PowerSystem,
    include_microgrid: bool = False,
    include_production: bool = False,
    data_dir=os.path.join(
        os.pardir,
        os.pardir,
        "data",
    ),
):

    # Fetching bus-objects
    B2 = power_system.get_comp("B2")
    B3 = power_system.get_comp("B3")
    B4 = power_system.get_comp("B4")
    B5 = power_system.get_comp("B5")
    B6 = power_system.get_comp("B6")
    B7 = power_system.get_comp("B7")
    B8 = power_system.get_comp("B8")
    B9 = power_system.get_comp("B9")
    B10 = power_system.get_comp("B10")

    if include_microgrid:
        M1 = power_system.get_comp("M1")
        M2 = power_system.get_comp("M2")
        M3 = power_system.get_comp("M3")
        # Fetching production objects
        P1 = M2.get_production()

    temp_profiles, wind_profiles, solar_profiles = weather_generation_data(
        path=os.path.join(data_dir, "weather_data_rygge.csv"),
    )

    wind = wind_power(wind_profiles)
    PV = pv_power(temp_profiles, solar_profiles)

    (
        load_household,
        load_farm,
        load_microgrid,
        load_industry2,
        load_trade,
        load_office,
    ) = load_data(
        temp_profiles,
        path=os.path.join(data_dir, "load_profiles_fasit.csv"),
    )

    microgrid = CostFunction(
        A=(21.4 - 17.5) * 1000,
        B=17.5 * 1000,
    )

    industry = CostFunction(
        A=132.6 - 92.5,
        B=92.5,
    )

    trade = CostFunction(
        A=220.3 - 102.4,
        B=102.4,
    )

    household = CostFunction(
        A=8.8,
        B=14.7,
    )

    for bus in [B3, B4, B6, B7, B9]:
        bus.add_load_data(
            pload_data=load_household,
            cost_function=household,
        )
    for bus in [B2, B8, B10]:
        bus.add_load_data(
            pload_data=load_industry2,
            cost_function=industry,
        )
    B5.add_load_data(
        pload_data=load_trade,
        cost_function=trade,
    )

    if include_microgrid:

        for bus in [M1, M2, M3]:
            bus.add_load_data(
                pload_data=load_microgrid,
                cost_function=microgrid,
            )

        if include_production:
            P1.add_prod_data(
                pprod_data=PV + wind,
            )
    return power_system



import matplotlib.lines as mlines
import numpy as np

from relsad.StatDist import StatDist, StatDistType, UniformParameters
from relsad.utils import convert_yearly_fail_rate, random_choice
from relsad.Time import Time

from .Component import Component


class ICTNode(Component):
    """
    Common base class for all ICT nodes

    ...

    Attributes
    ----------
    name : string
        Name of the ICT node
    coordinate : list
        Coordinate of the ICT node
    toline : Line
        Tells which line that is going into the ICT node
    fromline : Line
        Tells which line that is going out of the ICT node
    toline_list : list
        List of lines going into the ICT node
    fromline_list : list
        List of lines going from the ICT node
    connected_lines : List
        List of connected lines
    parent_network : ICTNetwork
        Parent ICT network of the ICT node
    fail_rate_per_year : float
        The failure rate per year for the transformer at the ICT node
    repair_time_dist : StatDist
        The repair time of the transformer at the ICT node [hours/fault]
    remaining_outage_time : Time
        The remaining outage time of the ICT node
    acc_outage_time : Time
        The accumulated outage time of the transformer at the ICT node
    avg_outage_time : Time
        The average outage time of the transformer at the ICT node
    avg_fail_rate : float
        The average failure rate of the transformer at the ICT node
    num_consecutive_interruptions : float
        Number of consecutive interruptions a ICT node experiences
    interruption_fraction : float
        The interruption fraction of the ICT node
    acc_interruptions : float
        Accumulated interruption duration a ICT node experiences
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict
        Dictonary attribute that stores the historic variables from the Monte Carlo simulation


    Methods
    ----------
    draw_repair_time(dt)
        Decides and returns the repair time of the trafo based on a statistical distribution
    fail(dt)
        Sets the ICT node status to failed
    not_fail()
        Sets the ICT node to not failed
    update_fail_status(dt)
        Updates the fail status of the ICT node. Sets the fail status to failed if the ICT node is failed or the fail status to not failed if the ICT node is not failed
    print_status()
        Prints the status of the ICT node
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute at the ICT node
    add_random_instance(random_gen)
        Adds global random seed
    get_avg_fail_rate(curr_time)
        Returns the average failure rate of the ICT node
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    get_monte_carlo_history(attribute)
        Returns a specified history variable from the Monte Carlo simulation
    """

    ## Visual attributes
    marker = "."
    size = 1.5**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=2,
        markersize=size,
        linestyle="None",
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        coordinate: list = [0, 0],
        fail_rate_per_year: float = 0.0,
        repair_time_dist: StatDist = StatDist(
            stat_dist_type=StatDistType.UNIFORM_FLOAT,
            parameters=UniformParameters(
                min_val=0.0,
                max_val=0.0,
            ),
        ),
    ):

        ## Informative attributes
        self.name = name
        self.coordinate = coordinate

        ## Topological attributes
        self.toline = None
        self.fromline = None
        self.toline_list = []
        self.fromline_list = []
        self.connected_lines = []
        self.parent_network = None

        ## Reliabilility attributes
        self.fail_rate_per_year = fail_rate_per_year  # failures per year
        self.repair_time_dist = repair_time_dist
        self.acc_outage_time = Time(0)
        self.avg_fail_rate = 0
        self.avg_outage_time = Time(0)
        self.num_consecutive_interruptions = 0
        self.interruption_fraction = 0
        self.acc_interruptions = 0

        ## Status attribute
        self.failed = False
        self.remaining_outage_time = Time(0)

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"ICTNode(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, ICTNode)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def draw_repair_time(self, dt: Time):
        """
        Decides and returns the repair time of the trafo based on a statistical distribution

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        return Time(
            self.repair_time_dist.draw(
                random_instance=self.ps_random,
                size=1,
            )[0],
            dt.unit,
        )

    def fail(self, dt: Time):
        """
        Sets the ICT node status to failed

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.failed = True
        self.remaining_outage_time = self.draw_repair_time(dt)

    def not_fail(self):
        """
        Sets the ICT node status to not failed

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = False

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the ICT node. Sets the fail status to failed if the ICT node is failed or the fail status to not failed if the ICT node is not failed

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.failed:
            self.remaining_outage_time -= dt
            if self.remaining_outage_time <= Time(0):
                self.not_fail()
                self.remaining_outage_time = Time(0)
        else:
            p_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
            if random_choice(self.ps_random, p_fail):
                self.fail(dt)
            else:
                self.not_fail()

    def print_status(self):
        """
        Prints the status of the ICT node

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        print(
            "name: {:3s}, failed={}".format(
                self.name,
                self.failed,
            )
        )

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["remaining_outage_time"] = {}
        self.history["failed"] = {}
        self.history["avg_fail_rate"] = {}
        self.history["avg_outage_time"] = {}
        self.history["acc_outage_time"] = {}
        self.history["acc_interruptions"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        dt = curr_time - prev_time if prev_time is not None else curr_time
        self.acc_outage_time += dt if self.failed is True else Time(0)
        self.avg_outage_time = Time(
            self.acc_outage_time / curr_time, curr_time.unit
        )
        self.avg_fail_rate = self.get_avg_fail_rate(curr_time)

        if self.failed is True:
            self.num_consecutive_interruptions += 1
        else:
            if self.num_consecutive_interruptions >= 1:
                self.acc_interruptions += 1
            self.num_consecutive_interruptions = 0

        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["remaining_outage_time"][
                time
            ] = self.remaining_outage_time.get_unit_quantity(curr_time.unit)
            self.history["failed"][time] = self.failed
            self.history["avg_fail_rate"][
                time
            ] = self.avg_fail_rate  # Average failure rate (lamda_s)
            self.history["avg_outage_time"][
                time
            ] = self.avg_outage_time.get_unit_quantity(
                curr_time.unit
            )  # Average outage time (r_s)
            self.history["acc_outage_time"][
                time
            ] = self.acc_outage_time.get_unit_quantity(
                curr_time.unit
            )  # Accumulated outage time
            self.history["acc_interruptions"][
                time
            ] = self.acc_interruptions

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute at the ICT node

        Parameters
        ----------
        attribute : str
            ICT node attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute
        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def get_avg_fail_rate(self, curr_time: Time):
        """
        Returns the average failure rate of the ICT node

        Parameters
        ----------
        None

        Returns
        ----------
        avg_fail_rate : float
            The average failure rate of the ICT node

        """
        fail_rate = self.fail_rate_per_year
        if self.parent_network is not None:
            for line in self.parent_network.get_lines():
                fail_rate += line.fail_rate_per_year
        avg_fail_rate = fail_rate / curr_time.get_years()
        return avg_fail_rate

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.failed = False
        self.remaining_outage_time = Time(0)
        self.acc_outage_time = Time(0)
        self.num_consecutive_interruptions = 0
        self.acc_interruptions = 0
        if save_flag:
            self.initialize_history()

    def get_monte_carlo_history(self, attribute):
        """
        Returns a specified history variable from the Monte Carlo simulation

        Parameters
        ---------
        attribute : str
            ICT node attribute

        Returns
        --------
        monte_carlo_history[attribute] : str
            The specified history variable from the Monte Carlo simulation
        """
        return self.monte_carlo_history[attribute]

    def get_neighbor_nodes(self):
        """
        Returns the neighboring nodes of the node

        Parameters
        ---------
        None

        Returns
        --------
        neighbor_nodes : list
            List of neighboring nodes
        """
        neighbor_nodes = []
        for line in self.connected_lines:
            if line.connected is True:
                for node in [line.fnode, line.tnode]:
                    if node not in neighbor_nodes and node != self:
                        neighbor_nodes.append(node)

        return neighbor_nodes



import os

import numpy as np

from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    EVPark,
    ICTLine,
    ICTNode,
    IntelligentSwitch,
    Line,
    MainController,
    ManualMainController,
    MicrogridMode,
    Production,
    Sensor,
)
from relsad.network.systems import (
    Distribution,
    ICTNetwork,
    Microgrid,
    PowerSystem,
    Transmission,
)
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network(
    fail_rate_trafo: float = 0.007,
    fail_rate_line: float = 0.07,
    fail_rate_intelligent_switch: float = 0.03,
    fail_rate_hardware: float = 0.2,
    fail_rate_software: float = 12,
    fail_rate_sensor: float = 0.023,
    fail_rate_ict_line: float = 0.068,
    p_fail_repair_new_signal: float = 1 - 0.95,
    p_fail_repair_reboot: float = 1 - 0.9,
    outage_time_trafo: Time = Time(8, TimeUnit.HOUR),
    include_microgrid: bool = True,
    include_production: bool = True,
    include_ICT: bool = True,
    include_ev: bool = True,
    v2g_flag: bool = True,
    ev_percentage: float = 0.47,
    ev_E_max: float = 0.07,
    include_backup: bool = True,
    microgrid_mode: MicrogridMode = MicrogridMode.SURVIVAL,
    line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
    ict_line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
):
    def num_ev_table_func(
        n_customers,
        ev_percentage=ev_percentage,
        daily_charge_frac=0.61,
    ):
        return Table(
            x=np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                ],
            ),
            y=np.array(
                [
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.08,
                    0.08,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.28,
                    0.28,
                    0.28,
                    0.28,
                    0.42,
                    0.42,
                    0.42,
                    0.42,
                ]
            )
            * n_customers
            * ev_percentage
            * daily_charge_frac,
        )

    if include_ICT:

        # Controller
        ICTNController = ICTNode(
            name="ICTNController",
            coordinate=[0, 2],
        )
        # Sensors
        ICTNSL1 = ICTNode(
            name="ICTNSL1",
        )
        ICTNSL2 = ICTNode(
            name="ICTNSL2",
        )
        ICTNSL3 = ICTNode(
            name="ICTNSL3",
        )
        ICTNSL4 = ICTNode(
            name="ICTNSL4",
        )
        ICTNSL5 = ICTNode(
            name="ICTNSL5",
        )
        ict_nodes = [
            ICTNController,
            ICTNSL1,
            ICTNSL2,
            ICTNSL3,
            ICTNSL4,
            ICTNSL5,
        ]
        if include_backup:
            ICTNSL6 = ICTNode(
                name="ICTNSL6",
            )
            ict_nodes.append(ICTNSL6)

        # Intelligent switch
        ICTNISW1a = ICTNode(
            name="ICTNISW1a",
        )
        ICTNISW1b = ICTNode(
            name="ICTNISW1b",
        )
        ICTNISW2a = ICTNode(
            name="ICTNISW2a",
        )
        ICTNISW2b = ICTNode(
            name="ICTNISW2b",
        )
        ICTNISW3a = ICTNode(
            name="ICTNISW3a",
        )
        ICTNISW3b = ICTNode(
            name="ICTNISW3b",
        )
        ICTNISW4a = ICTNode(
            name="ICTNISW4a",
        )
        ICTNISW4b = ICTNode(
            name="ICTNISW4b",
        )
        ICTNISW5a = ICTNode(
            name="ICTNISW5a",
        )
        ICTNISW5b = ICTNode(
            name="ICTNISW5b",
        )
        ict_nodes.extend(
            [
                ICTNISW1a,
                ICTNISW1b,
                ICTNISW2a,
                ICTNISW2b,
                ICTNISW3a,
                ICTNISW3b,
                ICTNISW4a,
                ICTNISW4b,
                ICTNISW5a,
                ICTNISW5b,
            ]
        )
        if include_backup:
            ICTNISW6a = ICTNode(
                name="ICTNISW6a",
            )
            ICTNISW6b = ICTNode(
                name="ICTNISW6b",
            )
            ict_nodes.extend(
                [
                    ICTNISW6a,
                    ICTNISW6b,
                ]
            )

        # Auxillary ICT nodes

        ## Access nodes
        ICTNA1 = ICTNode(
            name="ICTNA1",
            coordinate=[0.5, 2.25],
        )
        ICTNA2 = ICTNode(
            name="ICTNA2",
            coordinate=[0.5, 2.0],
        )
        ICTNA3 = ICTNode(
            name="ICTNA3",
            coordinate=[0.75, 2.0],
        )
        ICTNA4 = ICTNode(
            name="ICTNA4",
            coordinate=[0.75, 2.25],
        )
        ICTNA5 = ICTNode(
            name="ICTNA5",
            coordinate=[2.15, 1.5],
        )
        ICTNA6 = ICTNode(
            name="ICTNA6",
            coordinate=[2.4, 1.5],
        )
        ICTNA7 = ICTNode(
            name="ICTNA7",
            coordinate=[2.4, 1.75],
        )
        ICTNA8 = ICTNode(
            name="ICTNA8",
            coordinate=[2.15, 1.75],
        )
        ict_nodes.extend(
            [
                ICTNA1,
                ICTNA2,
                ICTNA3,
                ICTNA4,
                ICTNA5,
                ICTNA6,
                ICTNA7,
                ICTNA8,
            ]
        )

        ## Network nodes
        ICTN1 = ICTNode(
            name="ICTN1",
            coordinate=[1.0, 1.75],
        )
        ICTN2 = ICTNode(
            name="ICTN2",
            coordinate=[1.25, 1.5],
        )
        ICTN3 = ICTNode(
            name="ICTN3",
            coordinate=[1.5, 1.75],
        )
        ICTN4 = ICTNode(
            name="ICTN4",
            coordinate=[1.75, 2.0],
        )
        ICTN5 = ICTNode(
            name="ICTN5",
            coordinate=[2.25, 1.85],
        )
        ICTN6 = ICTNode(
            name="ICTN6",
            coordinate=[2.5, 2.0],
        )
        ICTN7 = ICTNode(
            name="ICTN7",
            coordinate=[2.5, 2.25],
        )
        ICTN8 = ICTNode(
            name="ICTN8",
            coordinate=[2.25, 2.5],
        )
        ICTN9 = ICTNode(
            name="ICTN9",
            coordinate=[1.75, 2.5],
        )
        ICTN10 = ICTNode(
            name="ICTN10",
            coordinate=[1.5, 2.0],
        )
        ICTN11 = ICTNode(
            name="ICTN11",
            coordinate=[1.5, 2.5],
        )
        ICTN12 = ICTNode(
            name="ICTN12",
            coordinate=[1.25, 2.25],
        )
        ict_nodes.extend(
            [
                ICTN1,
                ICTN2,
                ICTN3,
                ICTN4,
                ICTN5,
                ICTN6,
                ICTN7,
                ICTN8,
                ICTN9,
                ICTN10,
                ICTN11,
                ICTN12,
            ]
        )

        ## Lines

        ### Controller
        ICTL1 = ICTLine(
            name="ICTL1",
            fnode=ICTNController,
            tnode=ICTNA1,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )

        ### Sensors
        ICTL2 = ICTLine(
            name="ICTL2",
            fnode=ICTNSL1,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL3 = ICTLine(
            name="ICTL3",
            fnode=ICTNSL2,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL4 = ICTLine(
            name="ICTL4",
            fnode=ICTNSL3,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL5 = ICTLine(
            name="ICTL5",
            fnode=ICTNSL4,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL6 = ICTLine(
            name="ICTL6",
            fnode=ICTNSL5,
            tnode=ICTNA6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )

        ### Intelligent switches
        ICTL7 = ICTLine(
            name="ICTL7",
            fnode=ICTNISW1a,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL8 = ICTLine(
            name="ICTL8",
            fnode=ICTNISW1b,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL9 = ICTLine(
            name="ICTL9",
            fnode=ICTNISW2a,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL10 = ICTLine(
            name="ICTL10",
            fnode=ICTNISW2b,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL11 = ICTLine(
            name="ICTL11",
            fnode=ICTNISW3a,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL12 = ICTLine(
            name="ICTL12",
            fnode=ICTNISW3b,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL13 = ICTLine(
            name="ICTL13",
            fnode=ICTNISW4a,
            tnode=ICTNA6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL14 = ICTLine(
            name="ICTL14",
            fnode=ICTNISW4b,
            tnode=ICTNA5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL15 = ICTLine(
            name="ICTL15",
            fnode=ICTNISW5a,
            tnode=ICTNA6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL16 = ICTLine(
            name="ICTL16",
            fnode=ICTNISW5b,
            tnode=ICTNA6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ict_lines = [
            ICTL1,
            ICTL2,
            ICTL3,
            ICTL4,
            ICTL5,
            ICTL6,
            ICTL7,
            ICTL8,
            ICTL9,
            ICTL10,
            ICTL11,
            ICTL12,
            ICTL13,
            ICTL14,
            ICTL15,
            ICTL16,
        ]
        if include_backup:
            ICTL17 = ICTLine(
                name="ICTL17",
                fnode=ICTNSL6,
                tnode=ICTNA6,
                repair_time_dist=ict_line_repair_time_stat_dist,
                fail_rate_per_year=fail_rate_ict_line,
            )
            ICTL18 = ICTLine(
                name="ICTL18",
                fnode=ICTNISW6a,
                tnode=ICTNA6,
                repair_time_dist=ict_line_repair_time_stat_dist,
                fail_rate_per_year=fail_rate_ict_line,
            )
            ICTL19 = ICTLine(
                name="ICTL19",
                fnode=ICTNISW6b,
                tnode=ICTNA6,
                repair_time_dist=ict_line_repair_time_stat_dist,
                fail_rate_per_year=fail_rate_ict_line,
            )
            ict_lines.extend(
                [
                    ICTL17,
                    ICTL18,
                    ICTL19,
                ]
            )

        ### Access lines
        ICTL20 = ICTLine(
            name="ICTL20",
            fnode=ICTNA1,
            tnode=ICTNA2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL21 = ICTLine(
            name="ICTL21",
            fnode=ICTNA2,
            tnode=ICTNA3,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL22 = ICTLine(
            name="ICTL22",
            fnode=ICTNA1,
            tnode=ICTNA4,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL23 = ICTLine(
            name="ICTL23",
            fnode=ICTNA5,
            tnode=ICTNA6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL24 = ICTLine(
            name="ICTL24",
            fnode=ICTNA6,
            tnode=ICTNA7,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL25 = ICTLine(
            name="ICTL25",
            fnode=ICTNA5,
            tnode=ICTNA8,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ict_lines.extend(
            [
                ICTL20,
                ICTL21,
                ICTL22,
                ICTL23,
                ICTL24,
                ICTL25,
            ]
        )

        ### Network lines
        ICTL26 = ICTLine(
            name="ICTL26",
            fnode=ICTNA3,
            tnode=ICTN1,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL27 = ICTLine(
            name="ICTL27",
            fnode=ICTN1,
            tnode=ICTN2,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL28 = ICTLine(
            name="ICTL28",
            fnode=ICTN2,
            tnode=ICTNA8,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL29 = ICTLine(
            name="ICTL29",
            fnode=ICTN2,
            tnode=ICTN3,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL30 = ICTLine(
            name="ICTL30",
            fnode=ICTN2,
            tnode=ICTN10,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL31 = ICTLine(
            name="ICTL31",
            fnode=ICTN3,
            tnode=ICTN4,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL32 = ICTLine(
            name="ICTL32",
            fnode=ICTN4,
            tnode=ICTN5,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL33 = ICTLine(
            name="ICTL33",
            fnode=ICTN4,
            tnode=ICTN7,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL34 = ICTLine(
            name="ICTL34",
            fnode=ICTN4,
            tnode=ICTN8,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL35 = ICTLine(
            name="ICTL35",
            fnode=ICTN4,
            tnode=ICTN10,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL36 = ICTLine(
            name="ICTL36",
            fnode=ICTN5,
            tnode=ICTN6,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL37 = ICTLine(
            name="ICTL37",
            fnode=ICTN5,
            tnode=ICTNA7,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL38 = ICTLine(
            name="ICTL38",
            fnode=ICTN6,
            tnode=ICTN7,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL39 = ICTLine(
            name="ICTL39",
            fnode=ICTN8,
            tnode=ICTN9,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL40 = ICTLine(
            name="ICTL40",
            fnode=ICTN9,
            tnode=ICTN10,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL41 = ICTLine(
            name="ICTL41",
            fnode=ICTN9,
            tnode=ICTN11,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL42 = ICTLine(
            name="ICTL42",
            fnode=ICTN11,
            tnode=ICTN12,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ICTL43 = ICTLine(
            name="ICTL43",
            fnode=ICTN12,
            tnode=ICTNA4,
            repair_time_dist=ict_line_repair_time_stat_dist,
            fail_rate_per_year=fail_rate_ict_line,
        )
        ict_lines.extend(
            [
                ICTL26,
                ICTL27,
                ICTL28,
                ICTL29,
                ICTL30,
                ICTL31,
                ICTL32,
                ICTL33,
                ICTL34,
                ICTL35,
                ICTL36,
                ICTL37,
                ICTL38,
                ICTL39,
                ICTL40,
                ICTL41,
                ICTL42,
                ICTL43,
            ]
        )

    if include_ICT:
        C1 = MainController(
            name="C1",
            ict_node=ICTNController,
            hardware_fail_rate_per_year=fail_rate_hardware,
            software_fail_rate_per_year=fail_rate_software,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
    else:
        C1 = ManualMainController(
            name="C1",
            sectioning_time=Time(1, TimeUnit.HOUR),
        )

    ps = PowerSystem(C1)
    if include_ICT:
        ict_network = ICTNetwork(ps)
        ict_network.add_nodes(ict_nodes)
        ict_network.add_lines(ict_lines)

    ## Transmission network bus
    B1 = Bus(
        name="B1",
        n_customers=0,
        coordinate=[0, 0],
        fail_rate_per_year=0,
    )

    ## Distribution network
    B2 = Bus(
        name="B2",
        n_customers=1,
        coordinate=[1, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        name="B3",
        n_customers=100,
        coordinate=[2, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        name="B4",
        n_customers=50,
        coordinate=[2, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        name="B5",
        n_customers=90,
        coordinate=[3, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        name="B6",
        n_customers=3,
        coordinate=[3, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    L1 = Line(
        name="L1",
        fbus=B1,
        tbus=B2,
        r=0.057526629463617,
        x=0.029324854498807,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L2 = Line(
        name="L2",
        fbus=B2,
        tbus=B3,
        r=0.057526629463617,
        x=0.029324854498807,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L3 = Line(
        name="L3",
        fbus=B2,
        tbus=B4,
        r=0.7114,
        x=0.2351,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L4 = Line(
        name="L4",
        fbus=B4,
        tbus=B5,
        r=0.7114,
        x=0.2351,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )
    L5 = Line(
        name="L5",
        fbus=B3,
        tbus=B6,
        r=0.7114,
        x=0.2351,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_repair_time_stat_dist,
    )

    if include_backup:
        L6 = Line(
            name="L6",
            fbus=B4,
            tbus=B6,
            r=0.7114,
            x=0.2351,
            fail_rate_density_per_year=fail_rate_line,
            repair_time_dist=line_repair_time_stat_dist,
            capacity=6,
        )

    CircuitBreaker("E1", L1)

    DL1a = Disconnector("L1a", L1, B1)
    DL1b = Disconnector("L1b", L1, B2)
    DL2a = Disconnector("L2a", L2, B2)
    DL2b = Disconnector("L2b", L2, B3)
    DL3a = Disconnector("L3a", L3, B2)
    DL3b = Disconnector("L3b", L3, B4)
    DL4a = Disconnector("L4a", L4, B4)
    DL4b = Disconnector("L4b", L4, B5)
    DL5a = Disconnector("L5a", L5, B3)
    DL5b = Disconnector("L5b", L5, B6)

    if include_backup:
        DL6a = Disconnector("L6a", L6, B4)
        DL6b = Disconnector("L6b", L6, B6)

        L6.set_backup()

    if include_ICT:

        Sensor(
            name="SL1",
            line=L1,
            ict_node=ICTNSL1,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            name="SL2",
            line=L2,
            ict_node=ICTNSL2,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            name="SL3",
            line=L3,
            ict_node=ICTNSL3,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            name="SL4",
            line=L4,
            ict_node=ICTNSL4,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            name="SL5",
            line=L5,
            ict_node=ICTNSL5,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )

        if include_backup:
            Sensor(
                name="SL6",
                line=L6,
                ict_node=ICTNSL6,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

        IntelligentSwitch(
            name="RL1a",
            disconnector=DL1a,
            ict_node=ICTNISW1a,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL1b",
            disconnector=DL1b,
            ict_node=ICTNISW1b,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL2a",
            disconnector=DL2a,
            ict_node=ICTNISW2a,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL2b",
            disconnector=DL2b,
            ict_node=ICTNISW2b,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL3a",
            disconnector=DL3a,
            ict_node=ICTNISW3a,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL3b",
            disconnector=DL3b,
            ict_node=ICTNISW3b,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL4a",
            disconnector=DL4a,
            ict_node=ICTNISW4a,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL4b",
            disconnector=DL4b,
            ict_node=ICTNISW4b,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL5a",
            disconnector=DL5a,
            ict_node=ICTNISW5a,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )
        IntelligentSwitch(
            name="RL5b",
            disconnector=DL5b,
            ict_node=ICTNISW5b,
            fail_rate_per_year=fail_rate_intelligent_switch,
        )

        if include_backup:
            IntelligentSwitch(
                name="RL6a",
                disconnector=DL6a,
                ict_node=ICTNISW6a,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )
            IntelligentSwitch(
                name="RL6b",
                disconnector=DL6b,
                ict_node=ICTNISW6b,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )

    if include_ev:
        EVPark(
            name="EV1",
            bus=B5,
            num_ev_dist=num_ev_table_func(B3.n_customers),
            v2g_flag=v2g_flag,
            E_max=ev_E_max,
        )

    tn = Transmission(parent_network=ps, trafo_bus=B1)

    dn = Distribution(parent_network=tn, connected_line=L1)

    if include_microgrid:
        M1 = Bus(
            "M1",
            n_customers=0,
            coordinate=[2, -1],
            fail_rate_per_year=fail_rate_trafo,
        )
        M2 = Bus(
            "M2",
            n_customers=0,
            coordinate=[3, -2],
            fail_rate_per_year=fail_rate_trafo,
        )
        M3 = Bus(
            "M3",
            n_customers=40,
            coordinate=[3, -1],
            fail_rate_per_year=fail_rate_trafo,
        )

        Battery(name="Bat1", bus=M1)
        Production(name="P1", bus=M2)

        ML1 = Line(
            name="ML1",
            fbus=M1,
            tbus=M2,
            r=0.057526629463617,
            x=0.029324854498807,
            fail_rate_density_per_year=fail_rate_line,
            repair_time_dist=line_repair_time_stat_dist,
        )
        ML2 = Line(
            name="ML2",
            fbus=M1,
            tbus=M3,
            r=0.057526629463617,
            x=0.029324854498807,
            fail_rate_density_per_year=fail_rate_line,
            repair_time_dist=line_repair_time_stat_dist,
        )

        L7 = Line(
            name="L7",
            fbus=B2,
            tbus=M1,
            r=0.057526629463617,
            x=0.029324854498807,
            fail_rate_density_per_year=fail_rate_line,
            repair_time_dist=line_repair_time_stat_dist,
        )

        CircuitBreaker(name="E2", line=L7)

        DL7a = Disconnector(
            name="L7a",
            line=L7,
            bus=B2,
        )
        DL7b = Disconnector(
            name="L7b",
            line=L7,
            bus=M1,
        )
        DML1a = Disconnector(
            name="ML1a",
            line=ML1,
            bus=M1,
        )
        DML1b = Disconnector(
            name="ML1b",
            line=ML1,
            bus=M2,
        )
        DML2a = Disconnector(
            name="ML2a",
            line=ML2,
            bus=M1,
        )
        DML2b = Disconnector(
            name="ML2b",
            line=ML2,
            bus=M3,
        )

        if include_ICT:
            Sensor(
                name="SL7",
                line=L7,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

            Sensor(
                name="SML1",
                line=ML1,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )
            Sensor(
                name="SML2",
                line=ML2,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

            IntelligentSwitch(
                name="RL7a",
                disconnector=DL7a,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )
            IntelligentSwitch(
                name="RL7b",
                disconnector=DL7b,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )

            IntelligentSwitch(
                name="RML1a",
                disconnector=DML1a,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )
            IntelligentSwitch(
                name="RML1b",
                disconnector=DML1b,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )
            IntelligentSwitch(
                name="RML2a",
                disconnector=DML2a,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )
            IntelligentSwitch(
                name="RML2b",
                disconnector=DML2b,
                fail_rate_per_year=fail_rate_intelligent_switch,
            )

        m = Microgrid(
            distribution_network=dn,
            connected_line=L7,
            mode=microgrid_mode,
        )

        m.add_buses([M1, M2, M3])
        m.add_lines([ML1, ML2])

    dn.add_buses([B2, B3, B4, B5, B6])
    if include_backup:
        dn.add_lines([L2, L3, L4, L5, L6])
    else:
        dn.add_lines([L2, L3, L4, L5])

    return ps


if __name__ == "__main__":
    ps = initialize_network(
        include_microgrid=False,
        include_production=False,
        include_ICT=True,
        include_ev=False,
        include_backup=True,
    )

    fig = plot_topology(
        buses=ps.buses,
        lines=ps.lines,
        bus_text=True,
        line_text=True,
        ict_nodes=ps.ict_nodes,
        ict_lines=ps.ict_lines,
        text_dx=(0.05, -0.2),
        ncol=4,
    )

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "CINELDI.pdf",
        )
    )
    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "CINELDI.png",
        ),
        dpi=600,
    )



"""
This module contains comparison utilities
"""


def eq(x: float, y: float, tol: float = 1e-6):
    """
    Checks for equality within a tolerance

    Parameters
    ----------
    x : float
        Left hand side
    y : float
        Right hand side
    tol : float
        The equality tolerance

    Returns
    -------
    equal : bool
        Boolean variable stating whether the left and right hand side
        are equal within the given tolerance

    """
    equal = abs(x - y) < tol
    return equal



import numpy as np

from relsad.network.components import Bus, ICTLine, ICTNode, Line
from relsad.reliability.indices import (
    ASAI,
    ASUI,
    CAIDI,
    ENS,
    SAIDI,
    SAIFI,
    EV_Duration,
    EV_Index,
    EV_Interruption,
)
from relsad.Time import Time, TimeStamp
from relsad.topology.sectioning import create_sections, get_section_list
from relsad.utils import INF, eq, unique

from .PowerNetwork import PowerNetwork
from .Transmission import Transmission


class PowerSystem(PowerNetwork):
    """
    Class defining a power system type

    ...

    Attributes
    ----------
    name : str
        Name of the power system
    slack : Bus
        Slack bus of the power system
    p_energy_shed : float
        The active power energy.shed in the power system
    acc_p_energy_shed : float
        The accumulated active power energy.shedding in the power system
    q_energy_shed : float
        The reactive power energy.shed in the power system
    acc_q_energy_shed : float
        The accumulated reactive power energy.shedding in the power system
    sub_systems : list
        List of all sub systems in the power system
    buses : list
        List of all buses in the power system
    ev_parks : list
        List of all EV parks in the power system
    batteries : list
        List of all batteries in the power system
    production : list
        List of all generation units in the power system
    ict_nodes : list
        List of all ICT nodes in the power system
    lines : list
        List of all lines in the power system
    sensors : list
        List of all sensors in the power system
    circuitbreaker : list
        List of all circuit breakers in the power system
    disconnectors : list
        List of all disconnectors in the power system
    intelligent_switch : list
        List of all intelligent switches in the power system
    ict_lines : list
        List of all ICT lines in the power system
    controller : MainController
        The controller for the power system, either MainController or ManualMainController
    comp_list : list
        List containing the components in the power system
    comp_dict : dict
        Dictionary containing the components in the power system
    child_network_list : list
        List containing all the networks in the power system
    history : dict
        Dictionary containing the history variables of the power system
    monte_carlo_history : dict
        Dictionary containing the history variables from the monte carlo simulation



    Methods
    ----------
    add_bus(bus)
        Adding a bus including elements on the bus
        (battery, generation unit, EV park) to the power system
    add_buses(buses)
        Adding buses to the power system
    add_ict_node(ict_node)
        Adding an ICT node to the power system
    add_line(line)
        Adding a line including elements on the line
        (sensor, circuit breaker, disconnector) to the power system
    add_lines(lines)
        Adding lines to the power system
    add_ict_line(ict_line)
        Adding an ICT line to the power system
    get_lines()
        Returns the lines in the power system
    get_comp(name)
        Returns component based on given name
    get_comp_list()
        Returns list of the components in the power system
    add_child_network(network)
        Adding child network to the power system
    reset_slack_bus()
        Resets the slack bus of the child networks
    create_sections()
        Creates sections in the power system
    print_status()
        Prints the status of the buses and lines in the power system
    get_system_load_balance()
        Returns the load balance of the system in MW and MVar
    update_batteries(fail_duration, dt)
        Updates the batteries in the power system
    update_ev_parks(fail_duration, dt, start_time, curr_time)
        Updates the EV parks in the power system
    update_fail_status(dt)
        Updates the failure status for each component that
        can fail in the power system
    get_system_load()
        Returns the system load at the current time in MW and MVar
    prepare_load_data(time_indices)
        Prepares the load data for the buses in the power system
    prepare_prod_data(time_indices)
        Prepares the production data for the production components
        in the power system
    set_load_and_cost(inc_idx)
        Sets the bus load and cost in MW based on load and cost profiles
        in the current increment for the power system
    set_prod(inc_idx)
        Sets the generation (generation units, batteries, EV parks)
        at the buses in the power system
    failes_comp()
        Returns True if the power system contains a
        failed compoent, False otherwise
    full_batteries()
        Returns True if the batteries in the power system
        are full, and False otherwise
    reset_load_flow_data()
        Resets the variables used in the load flow analysis
    get_monte_carlo_history(attribute)
        Returns the specified history variable
        from the Monte Carlo simulation
    initialize_sequence_history()
        Initializes the dictionaries used for sequence history variables
    update_sequence_history()
        Updates the sequence history variables of the power system
    initialize_monte_carlo_history()
        Initializes the lists used for history variable
        from the Monte Carlo simulation
    update_monte_carlo_history()
        Updates the history dictionary from the Monte Carlo simulation
    update_monte_carlo_child_network_history()
        Updates the history dictionary for the child networks
        in the Monte Carlo simulation
    update_monte_carlo_comp_history()
        Updates the component values for the system
        from the Monte Carlo simulation
    get_history(attribute)
        Returns the specified history variable
    reset_energy_shed_variables()
        Resets the energy.shed variables
    verify_component_setup()
        Verifies the component setup in the power system

    """

    ## Visual attributes
    color = "black"

    ## Counter
    counter = 0

    ## Load shed configurations
    shed_configs: list = list()

    def __init__(self, controller):
        """Initializing power system content
        Content:
            buses(set): List of buses
            lines(set): List of lines
            comp_dict(dict): Dictionary of components
        """
        # Info
        PowerSystem.counter += 1
        self.name = "ps{:d}".format(PowerSystem.counter)
        # Load flow
        self.slack = None
        # Load shedding
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0
        # Sub-systems
        self.sub_systems: list = []
        # Components
        self.buses: list = []
        self.ev_parks: list = []
        self.batteries: list = []
        self.productions: list = []
        self.ict_nodes: list = []
        self.lines: list = []
        self.sensors: list = []
        self.circuitbreakers: list = []
        self.disconnectors: list = []
        self.intelligent_switches: list = []
        self.ict_lines: list = []
        self.controller = controller
        self.comp_list: list = []
        self.comp_dict: dict = {}
        ## Child networks
        self.child_network_list: list = []
        ## History
        self.history: dict = {}
        self.monte_carlo_history: dict = {}
        ## Random instance
        self.random_instance: np.random.Generator = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"PowerSystem(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "buses") and hasattr(other, "lines"):
            return set(unique(self.buses + self.lines)) == set(
                unique(other.buses + other.lines)
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_bus(self, bus: Bus):
        """
        Adding a bus including elements on the bus
        (battery, generation unit, EV park) to the power system

        Parameters
        ----------
        bus : Bus
            A bus element

        Returns
        ----------
        None

        """
        self.comp_dict[bus.name] = bus
        self.comp_list.append(bus)
        self.buses.append(bus)
        self.buses = unique(self.buses)
        if bus.ev_park is not None:
            self.comp_dict[bus.ev_park.name] = bus.ev_park
            self.comp_list.append(bus.ev_park)
            self.ev_parks.append(bus.ev_park)
            self.ev_parks = unique(self.ev_parks)
        if bus.battery is not None:
            self.comp_dict[bus.battery.name] = bus.battery
            self.comp_list.append(bus.battery)
            self.batteries.append(bus.battery)
            self.batteries = unique(self.batteries)
        if bus.prod is not None:
            self.comp_dict[bus.prod.name] = bus.prod
            self.comp_list.append(bus.prod)
            self.productions.append(bus.prod)
            self.productions = unique(self.productions)
        self.comp_list = unique(self.comp_list)

    def add_buses(self, buses: list):
        """
        Adding buses to the power system

        Parameters
        ----------
        buses : list
            A list of Bus elements in the power system

        Returns
        ----------
        None

        """
        for bus in buses:
            self.add_bus(bus)

    def add_ict_node(self, ict_node: ICTNode):
        """
        Adding an ICT node to the power system

        Parameters
        ----------
        ict_node : ICTNode
            An ICT node element

        Returns
        ----------
        None

        """
        self.comp_dict[ict_node.name] = ict_node
        self.comp_list.append(ict_node)
        self.ict_nodes.append(ict_node)
        self.ict_nodes = unique(self.ict_nodes)
        self.comp_list = unique(self.comp_list)

    def add_line(self, line: Line):
        """
        Adding a line including elements on the line
        (sensor, circuit breaker, disconnector) to the power system

        Parameters
        ----------
        line : Line
            A line element

        Returns
        ----------
        None

        """
        self.comp_dict[line.name] = line
        self.comp_list.append(line)
        self.lines.append(line)
        self.lines = unique(self.lines)
        if line.sensor:
            self.comp_dict[line.sensor.name] = line.sensor
            self.comp_list.append(line.sensor)
            self.sensors.append(line.sensor)
            self.sensors = unique(self.sensors)
        for discon in line.disconnectors:
            self.comp_dict[discon.name] = discon
            self.comp_list.append(discon)
            self.disconnectors.append(discon)
            self.disconnectors = unique(self.disconnectors)
            if discon.intelligent_switch:
                self.comp_dict[
                    discon.intelligent_switch.name
                ] = discon.intelligent_switch
                self.comp_list.append(discon.intelligent_switch)
                self.intelligent_switches.append(discon.intelligent_switch)
                self.intelligent_switches = unique(self.intelligent_switches)
        if line.circuitbreaker is not None:
            c_b = line.circuitbreaker
            self.comp_dict[c_b.name] = c_b
            self.comp_list.append(c_b)
            self.circuitbreakers.append(c_b)
            self.circuitbreakers = unique(self.circuitbreakers)
        self.comp_list = unique(self.comp_list)

    def add_lines(self, lines: list):
        """
        Adding lines to the power system

        Parameters
        ----------
        lines : list
            A list of Line elements in the power system

        Returns
        ----------
        None

        """
        for line in lines:
            self.add_line(line)

    def add_ict_line(self, ict_line: ICTLine):
        """
        Adding an ICT line to the power system

        Parameters
        ----------
        ict_line : ICTLine
            An ICT line element

        Returns
        ----------
        None

        """
        self.comp_dict[ict_line.name] = ict_line
        self.comp_list.append(ict_line)
        self.ict_lines.append(ict_line)
        self.ict_lines = unique(self.ict_lines)
        self.comp_list = unique(self.comp_list)

    def get_lines(self):
        """
        Returns the lines in the power system

        Parameters
        ----------
        lines : list
            A list of Line elements in the power system

        Returns
        ----------
        lines : list
            List of Line elements

        """
        return self.lines

    def get_comp(self, name: str):
        """
        Returns component based on given name

        Parameters
        ----------
        name : str
           Name of the component

        Returns
        ----------
        comp_dict[name] : dict
            Dictionary containing information about the component

        """
        try:
            return self.comp_dict[name]
        except KeyError:
            print(name)
            print("Component is not part of the network")
            exit()

    def get_comp_list(self):
        """
        Returns list of the components in the power system

        Parameters
        ----------
        None

        Returns
        ----------
        comp_list : list
            List of the components in the power system

        """
        return self.comp_list

    def add_child_network(self, network):
        """
        Adding child network to the power system

        Parameters
        ----------
        network : PowerNetwork
            The child network of the power system

        Returns
        ----------
        None

        """
        self.child_network_list.append(network)

    def reset_slack_bus(self):
        """
        Resets the slack bus of the child networks

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for child_network in self.child_network_list:
            child_network.reset_slack_bus()

    def create_sections(self):
        """
        Creates sections in the power system

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for network in self.child_network_list:
            if not isinstance(network, Transmission):
                parent_section = create_sections(network.connected_line)
                network.sections = get_section_list(parent_section, [])

    def print_status(self):
        """
        Prints the status of the buses and lines in the power system

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        print("Buses:")
        for bus in self.buses:
            bus.print_status()
        print("Lines:")
        for line in self.lines:
            line.print_status()

    def get_system_load_balance(self):
        """
        Returns the load balance of the system in MW and MVar

        Parameters
        ----------
        None

        Returns
        ----------
        system_load_balance_p : float
            The active power load balance in the power system (load - generation)
        system_load_balance_q : float
            The reactive power load balance in the power system (load - generation)

        """
        system_load_balance_p, system_load_balance_q = 0, 0
        for bus in self.buses:
            for child_network in self.child_network_list:
                if isinstance(child_network, Transmission):
                    if bus == child_network.get_trafo_bus():
                        system_load_balance_p = -INF
                        system_load_balance_q = 0
                        return system_load_balance_p, system_load_balance_q
            system_load_balance_p += bus.pload - bus.pprod
            system_load_balance_q += bus.qload - bus.qprod
        return system_load_balance_p, system_load_balance_q

    def update_batteries(self, fail_duration: Time, dt: Time):
        """
        Updates the batteries in the power system

        Parameters
        ----------
        fail_duration : Time
            The failure duration
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        p, q = self.get_system_load_balance()
        for battery in self.batteries:
            p, q = battery.update(p, q, fail_duration, dt)

    def update_ev_parks(
        self,
        fail_duration: Time,
        dt: Time,
        start_time: TimeStamp,
        curr_time: Time,
    ):
        """
        Updates the EV parks in the power system

        Parameters
        ----------
        fail_duration : Time
            The failure duration
        dt : Time
            The current time step
        start_time : TimeStamp
            Start time
        curr_time : Time
            The current time

        Returns
        ----------
        None

        """
        hour_of_day = start_time.get_hour_of_day(curr_time)
        p, q = self.get_system_load_balance()
        for ev_park in self.ev_parks:
            p, q = ev_park.update(
                p=p,
                q=q,
                fail_duration=fail_duration,
                dt=dt,
                hour_of_day=hour_of_day,
            )

    def update_fail_status(self, dt: Time):
        """
        Updates the failure status for each component that
        can fail in the power system

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.update_fail_status(dt)
        for battery in self.batteries:
            battery.update_fail_status(dt)
        for line in self.lines:
            line.update_fail_status(dt)
        for circuitbreaker in self.circuitbreakers:
            circuitbreaker.update_fail_status(dt)
        for sensor in self.sensors:
            sensor.update_fail_status(dt)
        for intelligent_switch in self.intelligent_switches:
            intelligent_switch.update_fail_status(dt)
        for ict_line in self.ict_lines:
            ict_line.update_fail_status(dt)
        for ict_node in self.ict_nodes:
            ict_node.update_fail_status(dt)
        self.controller.update_fail_status(dt)

    def get_system_load(self):
        """
        Returns the system load at the current time in MW and MVar

        Parameters
        ----------
        None

        Returns
        ----------
        pload : float
            The active power load in the power system
        qload : float
            The reactive power load in the power system

        """
        pload, qload = 0, 0
        for bus in self.buses:
            p, q = bus.get_load()
            pload += p
            qload += q
        return pload, qload

    def prepare_load_data(self, time_indices: np.ndarray):
        """
        Prepares the load data for the buses in the power system

        Parameters
        ----------
        time_indices : np.ndarray
            Time indices used to discretize the load data

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.prepare_load_data(time_indices)

    def prepare_prod_data(self, time_indices: np.ndarray):
        """
        Prepares the production data for the production components
        in the power system

        Parameters
        ----------
        time_indices : np.ndarray
            Time indices used to discretize the load data

        Returns
        ----------
        None

        """
        for prod in self.productions:
            prod.prepare_prod_data(time_indices)

    def set_load_and_cost(self, inc_idx: int):
        """
        Sets the bus load and cost in MW based on load and cost profiles
        in the current increment for the power system

        Parameters
        ----------
        inc_idx : int
            Increment index

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.set_load_and_cost(inc_idx)

    def set_prod(self, inc_idx: int):
        """
        Sets the generation (generation units, batteries, EV parks)
        at the buses in the power system

        Parameters
        ----------
        inc_idx : int
            Increment index

        Returns
        ----------
        None

        """
        for prod in self.productions:
            prod.set_prod(inc_idx)
        for bus in self.buses:
            if bus.battery is not None:
                bus.reset_prod()
            elif bus.ev_park is not None:
                bus.reset_prod()

    def failed_comp(self):
        """
        Returns True if the power system contains a
        failed component, False otherwise

        Parameters
        ----------
        None

        Returns
        ----------
        True/False

        """
        return any(
            bus.trafo_failed for bus in self.buses
        ) or any(line.failed for line in self.lines)

    def full_batteries(self):
        """
        Returns True if the batteries in the power system
        are full, and False otherwise

        Parameters
        ----------
        None

        Returns
        ----------
        True/False

        """
        return all(
            eq(battery.SOC, battery.SOC_max)
            for battery in self.batteries
        )

    def reset_load_flow_data(self):
        """
        Reset the variables used in the load flow analysis

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.reset_load_flow_data()
        for line in self.lines:
            line.reset_load_flow_data()

    def get_monte_carlo_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            Power system attribute

        Returns
        ----------
        monte_carlo_history[attribute] : dict
            Returns the history variables of an attribute from the Monte Carlo simulation

        """
        return self.monte_carlo_history[attribute]

    def initialize_sequence_history(self):
        """
        Initializes the dictionaries used for sequence history variables
    
        Parameters
        ----------
        None
    
        Returns
        ----------
        None
    
        """
        network_state_list = [
            "p_energy_shed",
            "q_energy_shed",
            "acc_p_energy_shed",
            "acc_q_energy_shed",
            "p_load",
            "q_load",
            "SAIFI",
            "SAIDI",
            "CAIDI",
            "ASAI",
            "ASUI",
            "ENS",
            "EV_Index",
            "EV_Interruption",
            "EV_Duration",
        ]
        for state_var in network_state_list:
            self.history[state_var] = {}
        for network in self.child_network_list:
            for state_var in network_state_list:
                network.history[state_var] = {}
        self.controller.initialize_history()

    def update_sequence_history(
        self,
        prev_time: Time,
        curr_time: Time,
        save_flag: bool,
    ):
        """
        Updates the sequence history variables of the power system

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            The current time
        save_flag : bool
            Indicates if saving is on/off

        Returns
        ----------
        None

        """
        time = curr_time.get_unit_quantity(curr_time.unit)
        for comp in self.comp_list:
            comp.update_history(prev_time, curr_time, save_flag)
        self.controller.update_history(prev_time, curr_time, save_flag)
        for network in self.child_network_list:
            for bus in network.buses:
                network.p_energy_shed += bus.p_energy_shed_stack
                network.q_energy_shed += bus.q_energy_shed_stack
            self.p_energy_shed += network.p_energy_shed
            self.q_energy_shed += network.q_energy_shed
            network.acc_p_energy_shed += network.p_energy_shed
            network.acc_q_energy_shed += network.q_energy_shed
        self.acc_p_energy_shed += self.p_energy_shed
        self.acc_q_energy_shed += self.q_energy_shed
        if save_flag:
            self_state_dict = {
                "p_energy_shed": self.p_energy_shed,
                "q_energy_shed": self.p_energy_shed,
                "acc_p_energy_shed": self.acc_p_energy_shed,
                "acc_q_energy_shed": self.acc_q_energy_shed,
                "p_load": self.get_system_load()[0],
                "q_load": self.get_system_load()[1],
                "SAIFI": SAIFI(self),
                "SAIDI": SAIDI(self),
                "CAIDI": CAIDI(self),
                "ASAI": ASAI(self, curr_time),
                "ASUI": ASUI(self, curr_time),
                "ENS": ENS(self),
                "EV_Index": EV_Index(self),
                "EV_Interruption": EV_Interruption(self),
                "EV_Duration": EV_Duration(self),
            }
            for state_var, value in self_state_dict.items():
                self.history[state_var][time] = value
            for network in self.child_network_list:
                network_state_dict = {
                    "p_energy_shed": network.p_energy_shed,
                    "q_energy_shed": network.q_energy_shed,
                    "acc_p_energy_shed": network.acc_p_energy_shed,
                    "acc_q_energy_shed": network.acc_q_energy_shed,
                    "p_load": network.get_system_load()[0],
                    "q_load": network.get_system_load()[1],
                    "SAIFI": SAIFI(network),
                    "SAIDI": SAIDI(network),
                    "CAIDI": CAIDI(network),
                    "ASAI": ASAI(network, curr_time),
                    "ASUI": ASUI(network, curr_time),
                    "ENS": ENS(network),
                    "EV_Index": EV_Index(network),
                    "EV_Interruption": EV_Interruption(network),
                    "EV_Duration": EV_Duration(network),
                }
                for state_var, value in network_state_dict.items():
                    network.history[state_var][time] = value
        self.p_energy_shed = 0
        self.q_energy_shed = 0
        for network in self.child_network_list:
            network.p_energy_shed = 0
            network.q_energy_shed = 0

    def initialize_monte_carlo_history(self):
        """
        Initializes the lists used for history variables
        from the Monte Carlo simulation
    
        Parameters
        ----------
        None
    
        Returns
        ----------
        save_dict : dict
            Dictionary with simulation results
    
        """
        network_state_list = [
            "acc_p_energy_shed",
            "acc_q_energy_shed",
            "SAIFI",
            "SAIDI",
            "CAIDI",
            "ASAI",
            "ASUI",
            "ENS",
            "EV_Index",
            "EV_Interruption",
            "EV_Duration",
        ]
        save_dict = {}
        save_dict[self.name] = {}
        for state_var in network_state_list:
            save_dict[self.name][state_var] = {}
        for network in self.child_network_list:
            save_dict[network.name] = {}
            for state_var in network_state_list:
                save_dict[network.name][state_var] = {}
        bus_state_list = [
            "acc_p_energy_shed",
            "acc_q_energy_shed",
            "avg_outage_time",
            "acc_outage_time",
            "interruption_fraction",
            "acc_interruptions",
        ]
        ev_park_state_list = [
            "acc_num_interruptions",
            "acc_exp_interruptions",
            "acc_exp_car_interruptions",
            "acc_interruption_duration",
            "acc_available_num_cars",
            "num_cars",
        ]
        for bus in self.buses:
            save_dict[bus.name] = {}
            for state_var in bus_state_list:
                save_dict[bus.name][state_var] = {}
            if bus.ev_park is not None:
                save_dict[bus.ev_park.name] = {}
                for state_var in ev_park_state_list:
                    save_dict[bus.ev_park.name][state_var] = {}
        return save_dict

    def update_monte_carlo_history(
        self,
        it: int,
        current_time: Time,
        save_dict: dict,
    ):
        """
        Updates the history dictionary from the Monte Carlo simulation
    
        Parameters
        ----------
        it : int
            The iteration number
        current_time : Time
            Current time
        save_dict : dict
            Dictionary with simulation results
    
        Returns
        ----------
        save_dict : dict
            Dictionary with simulation results
    
        """
        network_state_dict = {
            "acc_p_energy_shed": self.acc_p_energy_shed,
            "acc_q_energy_shed": self.acc_q_energy_shed,
            "SAIFI": SAIFI(self),
            "SAIDI": SAIDI(self),
            "CAIDI": CAIDI(self),
            "ASAI": ASAI(self, current_time),
            "ASUI": ASUI(self, current_time),
            "ENS": ENS(self),
            "EV_Index": EV_Index(self),
            "EV_Interruption": EV_Interruption(self),
            "EV_Duration": EV_Duration(self),
        }
        for state_var, value in network_state_dict.items():
            save_dict[self.name][state_var][it] = value
        save_dict = self.update_monte_carlo_child_network_history(
            it, current_time, save_dict
        )
        save_dict = self.update_monte_carlo_comp_history(it, save_dict)
        return save_dict
    
    def update_monte_carlo_child_network_history(
        self,
        it: int,
        current_time: Time,
        save_dict: dict,
    ):
        """
        Updates the history dictionary for the child networks
        in the Monte Carlo simulation
    
        Parameters
        ----------
        it : int
            The iteration number
        current_time : Time
            Current time
        save_dict : dict
            Dictionary with simulation results
    
        Returns
        ----------
        save_dict : dict
            Dictionary with simulation results
    
        """
        for network in self.child_network_list:
            network_state_dict = {
                "acc_p_energy_shed": network.acc_p_energy_shed,
                "acc_q_energy_shed": network.acc_q_energy_shed,
                "SAIFI": SAIFI(network),
                "SAIDI": SAIDI(network),
                "CAIDI": CAIDI(network),
                "ASAI": ASAI(network, current_time),
                "ASUI": ASUI(network, current_time),
                "ENS": ENS(network),
                "EV_Index": EV_Index(network),
                "EV_Interruption": EV_Interruption(network),
                "EV_Duration": EV_Duration(network),
            }
            for state_var, value in network_state_dict.items():
                save_dict[network.name][state_var][it] = value
        return save_dict
    
    def update_monte_carlo_comp_history(
        self,
        it: int,
        save_dict: dict,
    ):
        """
        Updates the component values for the system
        from the Monte Carlo simulation
    
        Parameters
        ----------
        it : int
            The iteration number
        save_dict : dict
            Dictionary with simulation results
    
        Returns
        ----------
        save_dict : dict
            Dictionary with simulation results
    
        """
        for bus in self.buses:
            bus_state_dict = {
                "acc_p_energy_shed": bus.acc_p_energy_shed,
                "acc_q_energy_shed": bus.acc_q_energy_shed,
                "avg_outage_time": bus.avg_outage_time.get_hours(),
                "acc_outage_time": bus.acc_outage_time.get_hours(),
                "interruption_fraction": bus.interruption_fraction,
                "acc_interruptions": bus.acc_interruptions,
            }
            for state_var, value in bus_state_dict.items():
                save_dict[bus.name][state_var][it] = value
            if bus.ev_park is not None:
                ev_park_state_dict = {
                    "acc_num_interruptions": bus.ev_park.acc_num_interruptions,
                    "acc_exp_interruptions": bus.ev_park.acc_exp_interruptions,
                    "acc_exp_car_interruptions": bus.ev_park.acc_exp_car_interruptions,
                    "acc_interruption_duration": bus.ev_park.acc_interruption_duration.get_hours(),
                    "acc_available_num_cars": bus.ev_park.acc_available_num_cars,
                    "num_cars": bus.ev_park.num_cars,
                }
                for state_var, value in ev_park_state_dict.items():
                    save_dict[bus.ev_park.name][state_var][it] = value
        return save_dict

    def get_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            Power system attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def reset_energy_shed_variables(self):
        """
        Resets the energy.shed variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0

    def verify_component_setup(self):
        """
        Verifies the component setup in the power system

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for bus in self.buses:
            for line in bus.connected_lines:
                if line not in self.lines and line.is_backup:
                    if line.parent_network is None:
                        raise Exception(
                            "Line, {:s}, does not have a parent network".format(
                                line.name
                            )
                        )



"""
This module contains random utilities
"""

import numpy as np

from relsad.Time import Time


def get_random_instance(seed: int = None):
    """
    Return a numpy random instance with optional seed

    Parameters
    ----------
    seed : int
        Random seed

    Returns
    -------
    random_instance : numpy.random.Generator
        Random instance generator

    """
    random_instance = np.random.default_rng(seed)
    return random_instance


def random_choice(random_instance: np.random.Generator, p_true: float):
    """
    Returns random choice based on uniform probability distribution

    Parameters
    ----------
    random_instance : np.random.Generator
        Instance of a random generator
    p_true : float
        The probability of true

    Returns
    -------
    choice : bool
        Random choice based on uniform probability distribution

    """
    if random_instance is None:
        random_instance = get_random_instance()
    choice = random_instance.random() < p_true
    return choice


def convert_yearly_fail_rate(fail_rate_per_year: float, dt: Time):
    """
    Returns the failure rate of the time increment, dt, based on
    the given yearly failure rate

    Parameters
    ----------
    fail_rate_per_year : float
        The failure rate
    dt : Time
        The current time step

    Returns
    -------
    fail_rate : float
        The converted failure rate for time increment, dt
    """
    fail_rate = min(fail_rate_per_year * dt.get_years(), 1)
    return fail_rate



from relsad.examples.tutorial.system_components import *

from relsad.network.systems import Distribution, PowerSystem

ps = PowerSystem(controller=C1)

dn = Distribution(
    parent_network=ps,
    connected_line=None,
)
dn.add_buses([B1, B2, B3, B4, B5, B6])
dn.add_lines([L1, L2, L3, L4, L5, L6])



"""
=============
Container types
=============

.. currentmodule:: relsad.network.containers


"""

from .Section import Section, SectionState

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



"""
==================
Network components
==================

.. currentmodule:: relsad.network.components


"""

from .Battery import Battery, BatteryState, BatteryType
from .Bus import Bus
from .CircuitBreaker import CircuitBreaker
from .Component import Component
from .Controller import Controller, ControllerState
from .Disconnector import Disconnector
from .DistributionController import DistributionController
from .EVPark import EVPark
from .ICTLine import ICTLine
from .ICTNode import ICTNode
from .IntelligentSwitch import IntelligentSwitch, IntelligentSwitchState
from .Line import Line
from .MainController import MainController
from .ManualMainController import ManualMainController
from .MicrogridController import MicrogridController, MicrogridMode
from .Production import Production
from .Sensor import Sensor, SensorState

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from relsad.network.components import Bus, Line
from relsad.Time import Time, TimeStamp, TimeUnit
from relsad.utils import INF, eq, unique

from .Transmission import Transmission


class SubSystem:
    """Class defining a sub system network type
    ...

    Attributes
    ----------
    name : str
        Name of the sub system
    slack : Bus
        The system slack bus, used for power flow calculations
    buses : list
        List of all buses in the sub system
    ev_parks : list
        List of all EV parks in the sub system
    batteries : list
        List of all batteries in the sub system
    production : list
        List of all generation units in the sub system
    lines : list
        List of all lines in the sub system
    sensors : list
        List of all sensors in the sub system
    circuitbreaker : list
        List of all circuit breakers in the sub system
    disconnectors : list
        List of all disconnectors in the sub system
    intelligent_switch : list
        List of all intelligent switches in the sub system
    comp_list : list
        List containing the components in the sub system
    comp_dict : dict
        Dictionary containing the components in the sub system
    child_network_list : list
        List containing the child networks to the sub system


    Methods
    ----------
    add_bus(bus)
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the sub system
    add_buses(buses)
        Adding buses to the sub system
    add_line(Line)
        Adding a line including elements on the line (sensor, circuit breaker, disconnector) to the sub system
    add_lines(lines)
        Adding lines to the sub system
    add_cild_network(network)
        Adding child network to the sub system
    get_system_load_balance()
        Returns the load balance of the system
    update_batteries(fail_duration, dt)
        Updates the batteries in the sub system
    update_ev_parks(fail_duration, dt, start_time, curr_time)
        Updates the EV parks in the sub system
    reset_load_flow_data()
        Reset the variables used in the load flow analysis

    """

    ## Visual attributes
    color = "black"

    ## Counter
    counter = 0

    def __init__(self):
        # Info
        SubSystem.counter += 1
        self.name = "ps{:d}".format(SubSystem.counter)
        # Load flow
        self.slack = None
        # Components
        self.buses = list()
        self.ev_parks = list()
        self.batteries = list()
        self.productions = list()
        self.lines = list()
        self.sensors = list()
        self.intelligent_switch = list()
        self.circuitbreakers = list()
        self.disconnectors = list()
        self.comp_list = list()
        self.comp_dict = dict()
        ## Child networks
        self.child_network_list: list = list()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"SubSystem(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "buses") and hasattr(other, "lines"):
            return set(unique(self.buses + self.lines)) == set(
                unique(other.buses + other.lines)
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_bus(self, bus: Bus):
        """
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the sub system

        Parameters
        ----------
        bus : Bus
            A bus element

        Returns
        ----------
        None

        """
        self.comp_dict[bus.name] = bus
        self.comp_list.append(bus)
        self.buses.append(bus)
        self.buses = unique(self.buses)
        if bus.ev_park is not None:
            self.comp_dict[bus.ev_park.name] = bus.ev_park
            self.comp_list.append(bus.ev_park)
            self.ev_parks.append(bus.ev_park)
            self.ev_parks = unique(self.ev_parks)
        if bus.battery is not None:
            self.comp_dict[bus.battery.name] = bus.battery
            self.comp_list.append(bus.battery)
            self.batteries.append(bus.battery)
            self.batteries = unique(self.batteries)
        if bus.prod is not None:
            self.comp_dict[bus.prod.name] = bus.prod
            self.comp_list.append(bus.prod)
            self.productions.append(bus.prod)
            self.productions = unique(self.productions)
        self.comp_list = unique(self.comp_list)

    def add_buses(self, buses: list):
        """Adding buses to the sub system

        Parameters
        ----------
        buses : list
            A list of Bus elements in the sub system

        Returns
        ----------
        None

        """
        for bus in buses:
            self.add_bus(bus)

    def add_line(self, line: Line):
        """
        Adding a line including elements on the line (sensor, circuit breaker, disconnector) to the sub system

        Parameters
        ----------
        line : Line
            A line element

        Returns
        ----------
        None

        """
        self.comp_dict[line.name] = line
        self.comp_list.append(line)
        self.lines.append(line)
        self.lines = unique(self.lines)
        for discon in line.disconnectors:
            self.comp_dict[discon.name] = discon
            self.comp_list.append(discon)
            self.disconnectors.append(discon)
            self.disconnectors = unique(self.disconnectors)
        if line.circuitbreaker is not None:
            c_b = line.circuitbreaker
            self.comp_dict[c_b.name] = c_b
            self.comp_list.append(c_b)
            self.circuitbreakers.append(c_b)
            self.circuitbreakers = unique(self.circuitbreakers)
        self.comp_list = unique(self.comp_list)

    def add_lines(self, lines: list):
        """
        Adding lines to the sub system

        Parameters
        ----------
        lines : list
            A list of Line elements in the sub system

        Returns
        ----------
        None

        """
        for line in lines:
            self.add_line(line)

    def add_child_network(self, network):
        """
        Adding child network to the sub system

        Parameters
        ----------
        network : PowerNetwork
            The child network of the sub system

        Returns
        ----------
        None

        """
        self.child_network_list.append(network)

    def get_system_load_balance(self):
        """
        Returns the load balance of the system

        Parameters
        ----------
        None

        Returns
        ----------
        system_load_balance_p : flaot
            The active power load balance in the sub system (load - generation)
        system_load_balance_q : float
            The reactive power load balance in the sub system (load - generation)

        """
        system_load_balance_p, system_load_balance_q = 0, 0
        for bus in self.buses:
            for child_network in self.child_network_list:
                if isinstance(child_network, Transmission):
                    if bus == child_network.get_trafo_bus():
                        system_load_balance_p = -INF
                        system_load_balance_q = 0
                        return system_load_balance_p, system_load_balance_q
            system_load_balance_p += bus.pload - bus.pprod
            system_load_balance_q += bus.qload - bus.qprod
        return system_load_balance_p, system_load_balance_q

    def update_batteries(self, fail_duration: Time, dt: Time):
        """
        Updates the batteries in the sub system

        Parameters
        ----------
        fail_duration : Time
            The failure duration
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        p, q = self.get_system_load_balance()
        for battery in self.batteries:
            p, q = battery.update(p, q, fail_duration, dt)

    def update_ev_parks(
        self,
        fail_duration: Time,
        dt: Time,
        start_time: TimeStamp,
        curr_time: Time,
    ):
        """
        Updates the EV parks in the sub system

        Parameters
        ----------
        fail_duration : Time
            The failure duration
        dt : Time
            The current time step
        start_time : TimeStamp
            Start time
        curr_time : Time
            The current time

        Returns
        ----------
        None

        """
        hour_of_day = start_time.get_hour_of_day(curr_time)
        p, q = self.get_system_load_balance()
        for ev_park in self.ev_parks:
            p, q = ev_park.update(
                p=p,
                q=q,
                fail_duration=fail_duration,
                dt=dt,
                hour_of_day=hour_of_day,
            )

    def reset_load_flow_data(self):
        """
        Reset the variables used in the load flow analysis

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.reset_load_flow_data()
        for line in self.lines:
            line.reset_load_flow_data()



from relsad.network.systems import PowerNetwork
from relsad.Time import Time
from relsad.utils import eq


def SAIFI(network: PowerNetwork):
    """
    Returns the current SAIFI (System average interruption failure index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    saifi : float
        The SAIFI value

    """
    interrupted_customers = sum(
        bus.acc_interruptions * bus.n_customers for bus in network.buses
    )
    total_customers = sum(bus.n_customers for bus in network.buses)
    if total_customers == 0:
        saifi = 0
    else:
        saifi = interrupted_customers / total_customers
    return saifi


def SAIDI(network: PowerNetwork):
    """
    Returns the current SAIDI (System average interruption duration index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    saidi : float
        The SAIDI value

    """
    sum_outage_time_x_n_customer = sum(
        bus.acc_outage_time.get_hours() * bus.n_customers
        for bus in network.buses
    )
    total_customers = sum(bus.n_customers for bus in network.buses)
    if total_customers == 0:
        saidi = 0
    else:
        saidi = sum_outage_time_x_n_customer / total_customers
    return saidi


def CAIDI(network: PowerNetwork):
    """
    Returns the current CAIFI (Customer average interruption duration index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    caidi : float
        The CAIDI value

    """
    saifi = SAIFI(network)
    if not eq(saifi, 0):
        caidi = SAIDI(network) / saifi
    else:
        caidi = 0
    return caidi


def ASUI(network: PowerNetwork, current_time: Time):
    """
    Returns the current ASUI (average service unavailability index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element
    current_time : Time
        Current time

    Returns
    ----------
    asui : float
        The ASUI value
    """
    hours = current_time.get_hours()
    asui = SAIDI(network) / hours
    return asui


def ASAI(network: PowerNetwork, current_time: Time):
    """
    Returns the current ASAI (average service avaialbility index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element
    current_time : Time
        Current time

    Returns
    ----------
    asai : float
        The ASAI value
    """
    asai = 1 - ASUI(network, current_time)
    return asai


def ENS(network):
    """
    Returns the current ENS (Energy not Supplied)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ens : float
        The ENS value

    """
    ens = sum(bus.acc_p_energy_shed for bus in network.buses)
    return ens



def flatten(toflatten: list):
    """
    Function that flattens nested list, handy for printing

    Parameters
    ----------
    toflatten : list
        Nested list

    Returns
    -------
    None

    """
    for element in toflatten:
        try:
            yield from flatten(element)
        except TypeError:
            yield element
